@@ -2,7 +2,7 @@ package Importer;
 use strict qw/vars subs/; # Not refs!
 use warnings; no warnings 'once';
 
-our $VERSION = '0.010';
+our $VERSION = '0.015';
 
 my %SIG_TO_SLOT = (
     '&' => 'CODE',
@@ -17,6 +17,8 @@ our %IMPORTED;
 # This will be used to check if an import arg is a version number
 my %NUMERIC = map +($_ => 1), 0 .. 9;
 
+sub IMPORTER_MENU() {(export_ok => [qw/optimal_import/])}
+
 ###########################################################################
 #
 # These are class methods
@@ -42,7 +44,7 @@ sub import {
     my $file = _mod_to_file($from);
     _load_file(\@caller, $file) unless $INC{$file};
 
-    return if _optimal_import($from, $caller[0], \@caller, @args);
+    return if optimal_import($from, $caller[0], \@caller, @args);
 
     my $self = $class->new(
         from   => $from,
@@ -81,7 +83,7 @@ sub import_into {
     my $file = _mod_to_file($from);
     _load_file(\@caller, $file) unless $INC{$file};
 
-    return if _optimal_import($from, $into, \@caller, @args);
+    return if optimal_import($from, $into, \@caller, @args);
 
     my $self = $class->new(
         from   => $from,
@@ -300,50 +302,58 @@ sub reload_menu {
 
     my $from = $self->from;
 
-    my (
-        $export,
-        $export_ok,
-        $export_tags,
-        $export_fail,
-        $generate,
-        $export_gen,
-        $export_anon,
-        $export_magic,
-        $new_style
-    );
-
     if (my $menu_sub = *{"$from\::IMPORTER_MENU"}{CODE}) {
         # Hook, other exporter modules can define this method to be compatible with
         # Importer.pm
 
-        $new_style = 1;
-
         my %got = $from->$menu_sub($into, $self->get_caller);
-        $export       = $got{export}       || [];
-        $export_ok    = $got{export_ok}    || [];
-        $export_tags  = $got{export_tags}  || {};
-        $export_fail  = $got{export_fail}  || [];
-        $export_anon  = $got{export_anon}  || {};
-        $export_magic = $got{export_magic} || {};
 
-        $export_gen = $got{export_gen};
-        $generate   = $got{generate};
+        $got{export}          ||= [];
+        $got{export_ok}       ||= [];
+        $got{export_tags}     ||= {};
+        $got{export_fail}     ||= [];
+        $got{export_anon}     ||= {};
+        $got{export_magic}    ||= {};
+        $got{export_versions} ||= {};
 
         $self->croak("'$from' provides both 'generate' and 'export_gen' in its IMPORTER_MENU (They are exclusive, module must pick 1)")
-            if $export_gen && $generate;
+            if $got{export_gen} && $got{generate};
 
-        $export_gen ||= {};
+        $got{export_gen} ||= {};
+
+        return $self->_build_menu($into => \%got, 1);
     }
     else {
-        $export       = \@{"$from\::EXPORT"};
-        $export_ok    = \@{"$from\::EXPORT_OK"};
-        $export_tags  = \%{"$from\::EXPORT_TAGS"};
-        $export_fail  = \@{"$from\::EXPORT_FAIL"};
-        $export_gen   = \%{"$from\::EXPORT_GEN"};
-        $export_anon  = \%{"$from\::EXPORT_ANON"};
-        $export_magic = \%{"$from\::EXPORT_MAGIC"};
-    }
+        my %got;
+        $got{export}          = \@{"$from\::EXPORT"};
+        $got{export_ok}       = \@{"$from\::EXPORT_OK"};
+        $got{export_tags}     = \%{"$from\::EXPORT_TAGS"};
+        $got{export_fail}     = \@{"$from\::EXPORT_FAIL"};
+        $got{export_gen}      = \%{"$from\::EXPORT_GEN"};
+        $got{export_anon}     = \%{"$from\::EXPORT_ANON"};
+        $got{export_magic}    = \%{"$from\::EXPORT_MAGIC"};
+        $got{export_versions} = \%{"$from\::EXPORT_VERSIONS"};
 
+        return $self->_build_menu($into => \%got, 0);
+    }
+}
+
+sub _build_menu {
+    my $self = shift;
+    my ($into, $got, $new_style) = @_;
+
+    my $from = $self->from;
+
+    my $export       = $got->{export}          || [];
+    my $export_ok    = $got->{export_ok}       || [];
+    my $export_tags  = $got->{export_tags}     || {};
+    my $export_fail  = $got->{export_fail}     || [];
+    my $export_anon  = $got->{export_anon}     || {};
+    my $export_magic = $got->{export_magic}    || {};
+    my $export_gen   = $got->{export_gen}      || {};
+    my $export_vers  = $got->{export_versions} || {};
+
+    my $generate = $got->{generate};
     $generate ||= sub {
         my $symbol = shift;
         my ($sig, $name) = ($symbol =~ m/^(\W?)(.*)$/);
@@ -398,6 +408,9 @@ sub reload_menu {
         'DEFAULT' => [ @$export ],
     };
 
+    # Add 'ALL' tag unless already specified. We want to normalize it.
+    $tags->{ALL} ||= [ sort grep {m/^[\&\$\@\%\*]/} keys %$lookup ];
+
     my $fail = @$export_fail ? {
         map {
             my ($sig, $name) = (m/^(\W?)(.*)$/);
@@ -406,6 +419,22 @@ sub reload_menu {
         } @$export_fail
     } : undef;
 
+    my $versions = {
+        v0 => { # Add base as v0, do not use the same hashref to avoid self-refrencing
+            lookup   => $lookup,
+            exports  => $exports,
+            tags     => $tags,
+            fail     => $fail,
+            generate => $generate,
+            magic    => $export_magic,
+        },
+    };
+    for my $v (keys %$export_vers) {
+        my $submenu = $self->_build_menu($into, $export_vers->{$v}, $new_style);
+        $tags->{$v} ||= [ "+$v", @{$submenu->{tags}->{DEFAULT}} ];
+        $versions->{$v} ||= $submenu;
+    }
+
     $self->{menu_for} = $into;
     return $self->{menu} = {
         lookup   => $lookup,
@@ -414,6 +443,7 @@ sub reload_menu {
         fail     => $fail,
         generate => $generate,
         magic    => $export_magic,
+        versions => $versions,
     };
 }
 
@@ -488,14 +518,21 @@ sub parse_args {
             my $pattern = $1;
             @list = sort grep /$1/, keys %{$menu->{lookup}};
         }
+        elsif($lead eq '+') {
+            my $vname = $arg;
+            substr($vname, 0, 1, '');
+
+            $self->croak("$from does not export the +$vname version")
+                unless $menu->{versions}->{$vname};
+
+            @list = ($arg);
+        }
         else {
             @list = ($arg);
         }
 
         # Normalize list, always have a sigil
-        my %seen;
-        @list = grep !$seen{$_}++, map {m/^\W/ ? $_ : "\&$_" } @list;
-
+        @list = map {m/^\W/ ? $_ : "\&$_" } @list;
 
         if ($exc) {
             $exclude{$_} = 1 for @list;
@@ -538,9 +575,9 @@ sub _set_symbols {
     my $self = shift;
     my ($into, $exclude, $import, $custom_set) = @_;
 
-    my $from   = $self->from;
-    my $menu   = $self->menu($into);
-    my $caller = $self->get_caller();
+    my $from      = $self->from;
+    my $main_menu = $self->menu($into);
+    my $caller    = $self->get_caller();
 
     my $set_symbol = $custom_set || eval <<"    EOT" || die $@;
 # Inherit the callers warning settings. If they have warnings and we
@@ -551,10 +588,17 @@ BEGIN { \${^WARNING_BITS} = \$caller->[9] if defined \$caller->[9] }
 sub { *{"$into\\::\$_[0]"} = \$_[1] }
     EOT
 
+    my $menu = $main_menu;
+    my $ver = "";
     for my $set (@$import) {
         my ($symbol, $spec) = @$set;
 
         my ($sig, $name) = ($symbol =~ m/^(\W)(.*)$/) or die "Invalid symbol: $symbol";
+        if ($sig eq '+') {
+            $ver = "version-set $name ";
+            $menu = $main_menu->{versions}->{$name};
+            next;
+        }
 
         # Find the thing we are actually shoving in a new namespace
         my $ref = $menu->{exports}->{$symbol};
@@ -563,7 +607,7 @@ sub { *{"$into\\::\$_[0]"} = \$_[1] }
         # Exporter.pm supported listing items in @EXPORT that are not actually
         # available for export. So if it is listed (lookup) but nothing is
         # there (!$ref) we simply skip it.
-        $self->croak("$from does not export $symbol") unless $ref || $menu->{lookup}->{"${sig}${name}"};
+        $self->croak("$from ${ver}does not export $symbol") unless $ref || $menu->{lookup}->{"${sig}${name}"};
         next unless $ref;
 
         my $type = ref($ref);
@@ -579,7 +623,7 @@ sub { *{"$into\\::\$_[0]"} = \$_[1] }
         my $new_name = join '' => ($spec->{'-prefix'} || '', $spec->{'-as'} || $name, $spec->{'-postfix'} || '');
 
         # Set the symbol (finally!)
-        $set_symbol->($new_name, $ref);
+        $set_symbol->($new_name, $ref, sig => $sig, symbol => $symbol, into => $into, from => $from, spec => $spec);
 
         # The remaining things get skipped with a custom setter
         next if $custom_set;
@@ -635,7 +679,7 @@ my %HEAVY_VARS = (
     EXPORT_MAGIC  => 'HASH',  # Origin package has magic to apply post-export
 );
 
-sub _optimal_import {
+sub optimal_import {
     my ($from, $into, $caller, @args) = @_;
 
     defined(*{"$from\::$_"}{$HEAVY_VARS{$_}}) and return 0 for keys %HEAVY_VARS;
@@ -691,7 +735,7 @@ L<Exporter>s specification. The exporter modules themselves do not need to use
 or inherit from the L<Exporter> module, they just need to set C<@EXPORT> and/or
 other variables.
 
-=head1 SYNOPSYS
+=head1 SYNOPSIS
 
     # Import defaults
     use Importer 'Some::Module';
@@ -733,7 +777,7 @@ addition to others.
 
 Adding a feature to L<Exporter> means that any consumer module that relies on
 the new features must depend on a specific version of L<Exporter>. This seems
-somewhat backwords since L<Exporter> is used by the module you are importing
+somewhat backwards since L<Exporter> is used by the module you are importing
 from.
 
 =item Exporter.pm is really old/crazy code
@@ -762,9 +806,9 @@ override this using the C<IMPORTER_MENU()> hook.
 
 =back
 
-=head1 COMPATABILITY
+=head1 COMPATIBILITY
 
-This module aims for 100% compatabilty with every feature of L<Exporter>, plus
+This module aims for 100% compatibility with every feature of L<Exporter>, plus
 added features such as import renaming.
 
 If you find something that works differently, or not at all when compared to
@@ -780,7 +824,7 @@ feature (like import renaming).
 =item $IMPORTER_VERSION (optional)
 
 If you provide a numeric argument as the first argument it will be treated as a
-version number. Importer will d a version check to make sure it is at least at
+version number. Importer will do a version check to make sure it is at least at
 the requested version.
 
 =item $FROM_MODULE (required)
@@ -812,10 +856,42 @@ implemented).
     $CARP{cluck}->("This will cluck");
     $CARP{croak}->("This will croak");
 
+The first two arguments to the custom sub are the name (no sigil), and the
+reference. The additional arguments are key/value pairs:
+
+    sub set_symbol {
+        my ($name, $ref, %info) = @_;
+    }
+
+=over 4
+
+=item $info{from}
+
+Package the symbol comes from.
+
+=item $info{into}
+
+Package to which the symbol should be added.
+
+=item $info{sig}
+
+The sigil that should be used.
+
+=item $info{spec}
+
+Extra details.
+
+=item $info{symbol}
+
+The original symbol name (with sigil) from the original package.
+
+=back
+
 =item @SYMBOLS (optional)
 
 Symbols you wish to import. If no symbols are specified then the defaults will
-be used.
+be used. You may also specify tags using the ':' prefix, or version-sets using
+the '+' symbol.
 
 =back
 
@@ -827,6 +903,54 @@ You can define/import subsets of symbols using predefined tags.
 
     use Importer 'Some::Thing' => ':tag';
 
+L<Importer> will automatically populate the C<:DEFAULT> tag for you.
+L<Importer> will also give you an C<:ALL> tag with ALL exports so long as the
+exporter does not define a C<:ALL> tag already.
+
+=head2 VERSION-SETS
+
+Some exporters may provide version-sets. Version sets are a way for exporters
+to provide alternate versions of exports. This is useful for maintaining
+backwards compatibility while providing a path forward.
+
+Importing without specifying a version set uses the default version set, which
+is also called 'v0'.
+
+You can specify an alternate version set with the '+' prefix:
+
+    use Importer 'Some::Thing' => qw/x +v1 a b c +v2 d +v0 y z/;
+
+The code above will import:
+
+=over 4
+
+=item x() from the v0 set (default set)
+
+Since no version was picked the default (v0) is used.
+
+=item a(), b(), and c() from the v1 set
+
+C<+v1> was specified, so the symbols requested after that are pulled from the
+v1 set.
+
+=item d() from the v2 set
+
+C<+v2> was specified, so the symbols requested after that are pulled from the
+v2 set.
+
+=item y() and z() from the v0 set
+
+C<+v0> was specified, so the symbols requested after that are pulled from the
+v0 set.
+
+=back
+
+Not all exporters provide versioned exports, but '+v0' is automatically
+generated and always present.
+
+See L</%EXPORT_VERSIONS> for details on providing version-sets from an
+exporter.
+
 =head2 /PATTERN/ or qr/PATTERN/
 
 You can import all symbols that match a pattern. The pattern can be supplied a
@@ -836,7 +960,7 @@ string starting and ending with '/', or you can provide a C<qr/../> reference.
 
     use Importer 'Some::Thing' => qr/oo/;
 
-=head2 EXLUDING SYMBOLS
+=head2 EXCLUDING SYMBOLS
 
 You can exclude symbols by prefixing them with '!'.
 
@@ -860,7 +984,7 @@ import name:
 You can also add a prefix and/or postfix:
 
     use Importer 'Some::Thing' => (
-        foo => { -prefix => 'my_foo' },
+        foo => { -prefix => 'my_' },
     );
 
 Using this syntax to set prefix and/or postfix also works on tags and patterns
@@ -869,7 +993,7 @@ all symbols from the tag/patterm.
 
 =head2 CUSTOM EXPORT ASSIGNMENT
 
-This lets you provide an alternative to the C<*name = $ref> export assingment.
+This lets you provide an alternative to the C<*name = $ref> export assignment.
 See the list of L<parameters|/"IMPORT PARAMETERS"> to C<import()>
 
 =head2 UNIMPORTING
@@ -963,12 +1087,13 @@ This allows you to export symbols that are generated on export. The key should
 be the name of a symbol. The value should be a coderef that produces a
 reference that will be exported.
 
-When the generators are called they will recieve 2 arguments, the package the
+When the generators are called they will receive 2 arguments, the package the
 symbol is being exported into, and the symbol being imported (name may or may
 not include sigil for subs).
 
     our %EXPORT_GEN = (
         '&foo' => sub {
+            my $from_package = shift;
             my ($into_package, $symbol_name) = @_;
             ...
             return sub { ... };
@@ -999,6 +1124,47 @@ custom assignment callback.
         },
     );
 
+=head2 %EXPORT_VERSIONS
+
+Export versions lets you provide different versions of exports potentially with
+the same name. This is a good way to maintain backwards compatibility while
+also providing a way forward if you have to make backwards incompatible
+changes.
+
+    package My::Thing;
+
+    our %EXPORT_ANON = (
+        foo => \&foo_orig,    # Export the original foo() implementation
+    );
+
+    our %EXPORT_VERSIONS = (
+        v1 => {
+            export_anon => {
+                foo => \&foo_v1,    # Export the v1 variant of foo()
+            },
+        },
+        latest => {
+            export => [qw/foo/],    # Export the latest implementation of foo()
+        },
+    );
+
+To use:
+
+This will import the original implementation
+
+    use Importer 'My::Thing' => qw/foo/;
+
+This will import the v1 variant
+
+    use Importer 'My::Thing' => qw/-v1 foo/;
+
+This will import the v1 latest foo()
+
+    use Importer 'My::Thing' => qw/-latest foo/;
+
+B<NOTE>: Nothing is inherited from the package variables/root menu. Anything
+you want in a specific version must be listed there.
+
 =head1 CLASS METHODS
 
 =over 4
@@ -1070,18 +1236,19 @@ B<IMPORTER_MENU() must be defined in your package, not a base class!>
         my ($into, $caller) = @_;
 
         return (
-            export       => \@EXPORT,          # Default exports
-            export_ok    => \@EXPORT_OK,       # Other allowed exports
-            export_tags  => \%EXPORT_TAGS,     # Define tags
-            export_fail  => \@EXPORT_FAIL,     # For subs that may not always be available
-            export_anon  => \%EXPORT_ANON,     # Anonymous symbols to export
-            export_magic => \%EXPORT_MAGIC,    # Magic to apply after a symbol is exported
+            export          => \@EXPORT,             # Default exports
+            export_ok       => \@EXPORT_OK,          # Other allowed exports
+            export_tags     => \%EXPORT_TAGS,        # Define tags
+            export_fail     => \@EXPORT_FAIL,        # For subs that may not always be available
+            export_anon     => \%EXPORT_ANON,        # Anonymous symbols to export
+            export_magic    => \%EXPORT_MAGIC,       # Magic to apply after a symbol is exported
+            export_versions => \%EXPORT_VERSIONS,    # Version sets to export
 
-            generate   => \&GENERATE,          # Sub to generate dynamic exports
-                                               # OR
-            export_gen => \%EXPORT_GEN,        # Hash of builders, key is symbol
-                                               # name, value is sub that generates
-                                               # the symbol ref.
+            generate   => \&GENERATE,                # Sub to generate dynamic exports
+                                                     # OR
+            export_gen => \%EXPORT_GEN,              # Hash of builders, key is symbol
+                                                     # name, value is sub that generates
+                                                     # the symbol ref.
         );
     }
 
@@ -1164,7 +1331,7 @@ This will remove imported symbols from the objects C<from> package. If you
 specify a list of C<@symbols> then only the specified symbols will be removed,
 otherwise all symbols imported using Importer will be removed.
 
-B<Note:> Please be aware fo the difference between C<do_import()> and
+B<Note:> Please be aware of the difference between C<do_import()> and
 C<do_unimport()>. For import 'from' us used as the origin, in unimport it is
 used as the target. This means you cannot re-use an instance to import and then
 unimport.
@@ -1262,10 +1429,31 @@ imports then only the LAST one will be used.
 
 =back
 
+=head1 FUNCTIONS
+
+These can be imported:
+
+    use Importer 'Importer' => qw/optimal_import/;
+
+=over 4
+
+=item $bool = optimal_import($from, $into, \@caller, @imports)
+
+This function will attempt to import C<@imports> from the C<$from> package into
+the C<$into> package. C<@caller> needs to have a package name, filename, and
+line number. If this function fails then no exporting will actually happen.
+
+If the import is successful this will return true.
+
+If the import is unsuccessful this will return false, and no modifications to
+the symbol table will occur.
+
+=back
+
 =head1 SOURCE
 
-The source code repository for symbol can be found at
-F<http://github.com/exodist/Importer>.
+The source code repository for Importer can be found at
+L<http://github.com/exodist/Importer>.
 
 =head1 MAINTAINERS
 
@@ -1290,6 +1478,6 @@ Copyright 2015 Chad Granum E<lt>exodist7@gmail.comE<gt>.
 This program is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself.
 
-See F<http://dev.perl.org/licenses/>
+See L<http://dev.perl.org/licenses/>
 
 =cut
